"""tests for vak.core.predict module"""
from pathlib import Path

import pandas as pd
import pytest

import vak.config
import vak.constants
import vak.core.predict


# written as separate function so we can re-use in tests/unit/test_cli/test_predict.py
def predict_output_matches_expected(output_dir, annot_csv_filename):
    annot_csv = output_dir.joinpath(annot_csv_filename)
    assert annot_csv.exists()

    return True


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format, save_net_outputs",
    [
        ("cbin", None, "notmat", False),
        ("wav", None, "birdsong-recognition-dataset", False),
        ("cbin", None, "notmat", True),
        ("wav", None, "birdsong-recognition-dataset", True),
    ],
)
def test_predict(
    audio_format,
    spect_format,
    annot_format,
    save_net_outputs,
    specific_config,
    tmp_path,
    model,
    device,
):
    output_dir = tmp_path.joinpath(
        f"test_predict_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "PREDICT", "option": "output_dir", "value": str(output_dir)},
        {"section": "PREDICT", "option": "device", "value": device},
        {"section": "PREDICT", "option": "save_net_outputs", "value": save_net_outputs},
    ]
    toml_path = specific_config(
        config_type="predict",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    model_config_map = vak.config.models.map_from_path(toml_path, cfg.predict.models)

    vak.core.predict(
        csv_path=cfg.predict.csv_path,
        checkpoint_path=cfg.predict.checkpoint_path,
        labelmap_path=cfg.predict.labelmap_path,
        model_config_map=model_config_map,
        window_size=cfg.dataloader.window_size,
        num_workers=cfg.predict.num_workers,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        spect_scaler_path=cfg.predict.spect_scaler_path,
        device=cfg.predict.device,
        annot_csv_filename=cfg.predict.annot_csv_filename,
        output_dir=cfg.predict.output_dir,
        min_segment_dur=cfg.predict.min_segment_dur,
        majority_vote=cfg.predict.majority_vote,
        save_net_outputs=cfg.predict.save_net_outputs,
        logger=None,
    )

    assert predict_output_matches_expected(output_dir, cfg.predict.annot_csv_filename)
    if save_net_outputs:
        net_outputs = sorted(
            Path(output_dir).glob(f"*{vak.constants.NET_OUTPUT_SUFFIX}")
        )

        vak_df = pd.read_csv(cfg.predict.csv_path)
        for spect_path in vak_df.spect_path.values:
            net_output_spect_path = [
                net_output
                for net_output in net_outputs
                if net_output.name.startswith(Path(spect_path).stem)
            ]
            assert len(net_output_spect_path) == 1



from .annot import *
from .audio import *
from .config import *
from .dataframe import *
from .device import *
from .model import *
from .path import *
from .spect import *
from .split import *
from .test_data import *
from .tensorboard import *
from .torch import *



"""fixtures for testing ``torch``-specific functionality"""
# adapted from kornia, https://github.com/kornia/kornia/blob/3606cf9c3d1eb3aabd65ca36a0e7cb98944c01ba/conftest.py
from typing import Dict

import pytest
import torch


def get_test_dtypes() -> Dict[str, torch.dtype]:
    """Creates a dictionary with the dtypes the source code.
    Return:
        dict(str, torch.dtype): list with dtype names.
    """
    dtypes: Dict[str, torch.dtype] = {}
    dtypes["float16"] = torch.float16
    dtypes["float32"] = torch.float32
    dtypes["float64"] = torch.float64
    return dtypes


TEST_DTYPES: Dict[str, torch.dtype] = get_test_dtypes()


@pytest.fixture()
def dtype(dtype_name) -> torch.dtype:
    return TEST_DTYPES[dtype_name]



"""fixtures relating to audio files"""
import pytest


@pytest.fixture
def default_spect_params():
    return dict(
        fft_size=512,
        step_size=64,
        freq_cutoffs=(500, 10000),
        thresh=6.25,
        transform_type="log_spect",
        freqbins_key="f",
        timebins_key="t",
        spect_key="s",
    )


@pytest.fixture
def audio_dir_cbin(source_test_data_root):
    return source_test_data_root.joinpath("audio_cbin_annot_notmat", "gy6or6", "032312")


@pytest.fixture
def audio_list_cbin(audio_dir_cbin):
    return sorted(audio_dir_cbin.glob("*.cbin"))


@pytest.fixture
def audio_list_cbin_all_labels_in_labelset(
    audio_list_cbin, annot_list_notmat, labelset_notmat
):
    """list of .cbin audio files where all labels in associated annotation **are** in labelset"""
    labelset_notmat = set(labelset_notmat)
    audio_list_labels_in_labelset = []
    for audio_path in audio_list_cbin:
        audio_fname = audio_path.name
        annot = [
            annot for annot in annot_list_notmat if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        if set(annot.seq.labels).issubset(labelset_notmat):
            audio_list_labels_in_labelset.append(audio_path)

    return audio_list_labels_in_labelset


@pytest.fixture
def audio_list_cbin_labels_not_in_labelset(
    audio_list_cbin, annot_list_notmat, labelset_notmat
):
    """list of .cbin audio files where some labels in associated annotation are **not** in labelset"""
    labelset_notmat = set(labelset_notmat)
    audio_list_labels_in_labelset = []
    for audio_path in audio_list_cbin:
        audio_fname = audio_path.name
        annot = [
            annot for annot in annot_list_notmat if annot.audio_path.name == audio_fname
        ]
        assert len(annot) == 1
        annot = annot[0]
        if not set(annot.seq.labels).issubset(labelset_notmat):
            audio_list_labels_in_labelset.append(audio_path)

    return audio_list_labels_in_labelset


# TODO: add .WAV from TIMIT
@pytest.fixture
def audio_dir_wav(source_test_data_root):
    return source_test_data_root.joinpath("audio_wav_annot_birdsongrec", "Bird0", "Wave")


@pytest.fixture
def audio_list_wav(audio_dir_wav):
    return sorted(audio_dir_wav.glob("*.wav"))


@pytest.fixture
def audio_list_factory(audio_list_cbin, audio_list_wav):
    """factory fixture, returns a function that
    returns a fixture containing a list of Annotation objects,
    given a specified annotation format

    so that unit tests can be parameterized with annotation format names
    """
    FORMAT_AUDIO_LIST_FIXTURE_MAP = {
        "cbin": audio_list_cbin,
        "wav": audio_list_wav,
    }

    def _audio_list_factory(format):
        return FORMAT_AUDIO_LIST_FIXTURE_MAP[format]

    return _audio_list_factory



from collections import defaultdict

import torch
import torch.nn.modules.loss
import torch.optim
from tqdm import tqdm

from ..device import get_default as get_default_device
from ..labeled_timebins import lbl_tb2labels
from ..logging import log_or_print


class Model:
    """lightweight model class that adds methods for training and evaluation
    to PyTorch neural networks.

    Sets forth a consistent API that models should implement to work with vak.

    Class Attributes
    ----------------
    network : torch.nn.Module
        subclass that implements a neural network
    loss : callable
        that implements a loss function, often called "criterion" in PyTorch code.
    optimizer : callable
        that optimizes model by iteratively updating trainable parameters.
    metrics : dict
        where keys are metric names, and values are callables that compute that a metric,
        e.g. accuracy. Metrics should accept arguments y_pred and y_true.

    Attributes
    ----------
    device : str
        device on which to place tensors. One of {"cuda", "cpu}.

    Methods
    -------
    fit : fit a model by training it with supplied data for a specified number of epochs
    evaluate : evaluate a model by computing specified metrics on supplied data
    predict : return predictions of model, i.e. output when fed with supplied data
    compile : returns instance of model with attributes set to specified arguments

    Private Methods
    ---------------
    _train : helper method, called by the fit method on each epoch.
        Fits the model by iterating once through all the training data.
        Override this method if you need to implement your own training method.
    _eval : helper method, called by the evaluate method, and called by the fit
        method for validation after each epoch. Evaluates the model by iterating
        through eval_data and computing the specified metrics for each batch.
        Override this method if you need to implement your own validation and test method.
    _predict : helper method, called by the predict method on each epoch.
        Uses the model to make predictions, by iterating through pred_data
        and returning the outputs of each batch fed into it.
        Override this method if you need to implement your own predict method.
    """

    REQUIRED_SUBCLASS_ATTRIBUTES = [
        "network",
        "optimizer",
        "loss",
        "metrics",
    ]

    def __init__(
        self,
        network,
        loss,
        optimizer,
        metrics,
        logger=None,
        summary_writer=None,
        global_step=0,
    ):
        self.network = network
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics

        self.logger = logger
        self.summary_writer = summary_writer
        self.global_step = global_step  # used for summary writer

        # attributes set by fit / _train methods
        self.device = None
        self.ckpt_path = None
        self.max_val_acc = 0
        self.max_val_acc_ckpt_path = None
        self.patience = None
        self.patience_counter = 0

    def _train(
        self,
        train_data,
        epoch,
        val_data=None,
        val_step=None,
        ckpt_step=None,
    ):
        """helper method, called by the fit method on each epoch.
        Iterates once through train_data, using it to update model parameters.
        Override this method if you need to implement your own training method.

        Parameters
        ----------
        train_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        self.network.train()

        progress_bar = tqdm(train_data)
        for ind, batch in enumerate(progress_bar):
            x, y = batch[0].to(self.device), batch[1].to(self.device)
            y_pred = self.network.forward(x)
            self.optimizer.zero_grad()
            loss = self.loss(y_pred, y)
            loss.backward()
            self.optimizer.step()
            progress_bar.set_description(
                f"Epoch {epoch}, batch {ind}. Loss: {loss.item():.4f}. Global step: {self.global_step}"
            )

            if self.summary_writer is not None:
                self.summary_writer.add_scalar(
                    "loss/train", loss.item(), self.global_step
                )
            self.global_step += 1

            if val_data is not None:
                if self.global_step % val_step == 0:
                    log_or_print(
                        f"Step {self.global_step} is a validation step; computing metrics on validation set",
                        logger=self.logger,
                        level="info",
                    )
                    metric_vals = self._eval(val_data)
                    self.network.train()  # because _eval calls network.eval()
                    log_or_print(
                        msg=", ".join(
                            [
                                f"{metric_name}: {metric_value:.4f}"
                                for metric_name, metric_value in metric_vals.items()
                                if metric_name.startswith("avg_")
                            ]
                        ),
                        logger=self.logger,
                        level="info",
                    )

                    if self.summary_writer is not None:
                        for metric_name, metric_value in metric_vals.items():
                            if metric_name.startswith("avg_"):
                                self.summary_writer.add_scalar(
                                    f"{metric_name}/val", metric_value, self.global_step
                                )

                    current_val_acc = metric_vals["avg_acc"]
                    if current_val_acc > self.max_val_acc:
                        self.max_val_acc = current_val_acc
                        log_or_print(
                            msg=f"Accuracy on validation set improved. Saving max-val-acc checkpoint.",
                            logger=self.logger,
                            level="info",
                        )
                        self.save(
                            self.max_val_acc_ckpt_path,
                            epoch=epoch,
                            global_step=self.global_step,
                        )
                        if self.patience:
                            self.patience_counter = 0
                    else:  # if accuracy did not improve
                        if self.patience:
                            self.patience_counter += 1
                            if self.patience_counter > self.patience:
                                log_or_print(
                                    "Stopping training early, "
                                    f"accuracy has not improved in {self.patience} validation steps.",
                                    logger=self.logger,
                                    level="info",
                                )
                                # save "backup" checkpoint upon stopping; don't save over "max-val-acc" checkpoint
                                self.save(
                                    self.ckpt_path,
                                    epoch=epoch,
                                    global_step=self.global_step,
                                )
                                progress_bar.close()
                                break
                            else:
                                log_or_print(
                                    f"Accuracy has not improved in {self.patience_counter} validation steps. "
                                    f"Not saving max-val-acc checkpoint for this validation step.",
                                    logger=self.logger,
                                    level="info",
                                )
                        else:  # patience is None. We still log that we are not saving checkpoint.
                            log_or_print(
                                "Accuracy is less than maximum validation accuracy so far. "
                                "Not saving max-val-acc checkpoint.",
                                logger=self.logger,
                                level="info",
                            )

            # below can be true regardless of whether we have val_data and/or current epoch is a val_epoch
            if self.global_step % ckpt_step == 0:
                log_or_print(
                    f"Step {self.global_step} is a checkpoint step.",
                    logger=self.logger,
                    level="info",
                )
                self.save(self.ckpt_path, epoch=epoch, global_step=self.global_step)

    def _eval(self, eval_data):
        """helper method, called by the evaluate method, and called by the fit
        method for validation after each epoch. Evaluates the model by iterating
        through eval_data and computing the specified metrics for each batch.
        Override this method if you need to implement your own validation and test method.

        Parameters
        ----------
        eval_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        self.network.eval()

        metric_vals = defaultdict(list)

        n_batches = 0

        progress_bar = tqdm(eval_data)
        with torch.no_grad():
            for ind, batch in enumerate(progress_bar):
                x, y = batch["source"].to(self.device), batch["annot"].to(self.device)
                # remove "batch" dimension added by collate_fn to x
                # we keep for y because loss still expects the first dimension to be batch
                if x.ndim == 5:
                    if x.shape[0] == 1:
                        x = torch.squeeze(x, dim=0)
                else:
                    raise ValueError(f"invalid shape for x: {x.shape}")

                out = self.network.forward(x)
                # permute and flatten out
                # so that it has shape (1, number classes, number of time bins)
                # ** NOTICE ** just calling out.reshape(1, out.shape(1), -1) does not work, it will change the data
                out = out.permute(1, 0, 2)
                out = torch.flatten(out, start_dim=1)
                out = torch.unsqueeze(out, dim=0)
                # reduce to predictions, assuming class dimension is 1
                y_pred = torch.argmax(
                    out, dim=1
                )  # y_pred has dims (batch size 1, predicted label per time bin)

                if "padding_mask" in batch:
                    padding_mask = batch[
                        "padding_mask"
                    ]  # boolean: 1 where valid, 0 where padding
                    # remove "batch" dimension added by collate_fn
                    # because this extra dimension just makes it confusing to use the mask as indices
                    if padding_mask.ndim == 2:
                        if padding_mask.shape[0] == 1:
                            padding_mask = torch.squeeze(padding_mask, dim=0)
                    else:
                        raise ValueError(
                            f"invalid shape for padding mask: {padding_mask.shape}"
                        )

                    out = out[:, :, padding_mask]
                    y_pred = y_pred[:, padding_mask]

                if any(
                    [
                        "levenshtein" in metric_name
                        for metric_name in self.metrics.keys()
                    ]
                ) or any(
                    [
                        "segment_error_rate" in metric_name
                        for metric_name in self.metrics.keys()
                    ]
                ):
                    y_labels = lbl_tb2labels(
                        y.cpu().numpy(), eval_data.dataset.labelmap
                    )
                    y_pred_labels = lbl_tb2labels(
                        y_pred.cpu().numpy(), eval_data.dataset.labelmap
                    )
                else:
                    y_labels = None
                    y_pred_labels = None

                for metric_name, metric_callable in self.metrics.items():
                    if metric_name == "loss":
                        metric_vals[metric_name].append(metric_callable(out, y))
                    elif metric_name == "acc":
                        metric_vals[metric_name].append(metric_callable(y_pred, y))
                    elif metric_name == "levenshtein":
                        metric_vals[metric_name].append(
                            metric_callable(y_pred_labels, y_labels)
                        )
                    elif metric_name == "segment_error_rate":
                        metric_vals[metric_name].append(
                            metric_callable(y_pred_labels, y_labels)
                        )
                    else:
                        raise NotImplementedError(
                            f"calculation of metric not yet implemented for {metric_name}"
                        )

                n_batches += 1
                progress_bar.set_description(f"batch {ind} / {len(eval_data)}")

        # ---- compute metrics averaged across batches -----------------------------------------------------------------
        # iterate over list of keys, to avoid "dictionary changed size" error when adding average metrics
        for metric_name in list(metric_vals):
            if metric_name in ["loss", "acc", "levenshtein", "segment_error_rate"]:
                avg_metric_val = (
                    torch.tensor(metric_vals[metric_name]).sum().cpu().numpy()
                    / n_batches
                ).item()
                metric_vals[f"avg_{metric_name}"] = avg_metric_val
            else:
                raise NotImplementedError(
                    f"calculation of metric across batches not yet implemented for {metric_name}"
                )

        return metric_vals

    def _predict(self, pred_data):
        """helper method, called by the predict method.
        Uses the model to make predictions, by iterating through pred_data
        and returning the outputs of each batch fed into it.
        Override this method if you need to implement your own predict method.

        Parameters
        ----------
        pred_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        preds = {}
        self.network.eval()

        progress_bar = tqdm(pred_data)

        with torch.no_grad():
            for ind, batch in enumerate(progress_bar):
                x, spect_path = batch["source"].to(self.device), batch["spect_path"]
                if isinstance(spect_path, list) and len(spect_path) == 1:
                    spect_path = spect_path[0]
                if x.ndim == 5:
                    if x.shape[0] == 1:
                        x = torch.squeeze(x, dim=0)
                y_pred = self.network.forward(x)
                preds[spect_path] = y_pred
                progress_bar.set_description(f"batch {ind} / {len(pred_data)}")

        return preds

    def save(self, ckpt_path, **kwargs):
        """save model state to a checkpoint file.

        Saves network state_dict, optimizer state_dict, and
        any keyword arguments to a checkpoint file with the name
        specified by ckpt_path.

        Parameters
        ----------
        ckpt_path : str, Path
            path including filename that should be used to save checkpoint
        kwargs :
            keyword arguments; if there are any, they will be added to checkpoint
        """
        ckpt = {
            "network_state_dict": self.network.state_dict(),
            "optimizer_state_dict": self.optimizer.state_dict(),
        }
        ckpt.update(**kwargs)
        log_or_print(
            f"Saving checkpoint at:\n{ckpt_path} ", logger=self.logger, level="info"
        )
        torch.save(ckpt, ckpt_path)

    def load(self, ckpt_path, device=None):
        """load model state dict from a checkpoint file.

        Loads state_dicts into network and optimizer
        from a checkpoint file with the name
        specified by ckpt_path.

        Parameters
        ----------
        ckpt_path : str, Path
            path including filename from which to load checkpoint
        """
        if device is None:
            device = get_default_device()

        log_or_print(
            f"Loading checkpoint from:\n{ckpt_path} ", logger=self.logger, level="info"
        )
        ckpt = torch.load(ckpt_path, map_location=device)
        self.network.load_state_dict(ckpt["network_state_dict"])
        self.optimizer.load_state_dict(ckpt["optimizer_state_dict"])

    def fit(
        self,
        train_data,
        num_epochs,
        ckpt_root,
        val_data=None,
        val_step=None,
        ckpt_step=None,
        patience=None,
        device=None,
    ):
        # ---- pre-conditions ----------
        if val_data is None:
            if patience is not None:
                raise ValueError(
                    f"patience set to {patience}, but no validation dataset was provided to measure accuracy"
                )
            if val_step is not None:
                raise ValueError(
                    f"val_step set to {val_step}, but no validation dataset was provided to measure accuracy"
                )

        # ---- set attributes ----------
        if device is None:
            device = get_default_device()
        self.device = device

        # note there can be up to two checkpoint paths.
        # this first one is the "backup" checkpoint, saved intermittently (with frequency determined by ckpt_step)
        # and also saved at the end of training
        self.ckpt_path = ckpt_root.joinpath("checkpoint.pt")

        if val_data is not None:
            # this is the second checkpoint path, saved when accuracy improves on the validation set
            self.max_val_acc_ckpt_path = ckpt_root.joinpath("max-val-acc-checkpoint.pt")
            self.max_val_acc = 0

        if patience is not None:
            self.patience = patience
            self.patience_counter = 0

        self.network.to(self.device)

        # ---- actually do fitting ----------
        for epoch in range(1, num_epochs + 1):
            log_or_print(
                f"epoch {epoch} / {num_epochs}", logger=self.logger, level="info"
            )
            self._train(train_data, epoch, val_data, val_step, ckpt_step)
            if patience is not None:
                if self.patience_counter > self.patience:
                    # need to break here too, not just inside _train function
                    break

        if (
            epoch == num_epochs
        ):  # save at end, if we complete all epochs (not if we stopped because of patience)
            log_or_print("Completed last epoch.", logger=self.logger, level="info")
            self.save(self.ckpt_path, epoch=epoch, global_step=self.global_step)

    def evaluate(self, eval_data, device=None):
        if device is None:
            device = get_default_device()
        self.device = device
        self.network.to(self.device)
        return self._eval(eval_data)

    def predict(self, pred_data, device=None):
        if device is None:
            device = get_default_device()
        self.device = device
        self.network.to(self.device)
        return self._predict(pred_data)

    @classmethod
    def from_config(cls, config, logger=None):
        """any model that inherits from this class should do whatever it needs to
        in this factory method to create the network, optimizer, and loss, and
        then pass those to the init function

        Parameters
        ----------
        config : dict
            presumably mapping 'network', 'optimizer' and 'loss' to kwargs that
            determine parameters for each of those, e.g. dropout rate for the
            network, learning rate for the optimizer, etc.
        logger : logging.Logger
            instance returned by vak.logging.get_logger.
            Default is None, in which case messages are just sent to print function.

        Returns
        -------
        model : Model
            instance of a model that subclasses this Model
        """
        raise NotImplementedError



def validate_split_durations(train_dur, val_dur, test_dur, dataset_dur):
    """helper function to validate durations specified for splits,
    so other functions can do the actual splitting.

    First the functions checks for invalid conditions:
        + If train_dur, val_dur, and test_dur are all None, a ValueError is raised.
        + If any of train_dur, val_dur, or test_dur have a negative value that is not -1, an
          ValueError is raised. -1 is interpreted differently as explained below.
        + If only val_dur is specified, this raises a ValueError; not clear what durations of training
          and test set should be.

    Then, if either train_dur or test_dur are None, they are set to 0. None means user did not specify a value.

    Finally the function validates that the sum of the specified split durations is not greater than
    the the total duration of the dataset, `dataset_dur`.

    If any split is specified as -1, this value is interpreted as "first get the
    split for the set with a value specified, then use the remainder of the dataset in the split
    whose duration is set to -1". Functions that do the splitting have to "know"
    about this meaning of -1, so this validation function does not modify the value.

    Parameters
    ----------
    train_dur : int, float
        Target duration for training set split, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    dataset_dur : int, float
        Total duration of dataset of vocalizations that will be split.

    Returns
    -------
    train_dur, val_dur, test_dur : int, float
    """
    if val_dur and (train_dur is None and test_dur is None):
        raise ValueError(
            "cannot specify only val_dur, unclear how to split dataset into training and test sets"
        )

    # make a dict so we can refer to variable by name in loop
    split_durs = {
        "train": train_dur,
        "val": val_dur,
        "test": test_dur,
    }
    if all([dur is None for dur in split_durs.values()]):
        raise ValueError(
            "train_dur, val_dur, and test_dur were all None; must specify at least train_dur or test_dur"
        )

    if not all(
        [dur >= 0 or dur == -1 for dur in split_durs.values() if dur is not None]
    ):
        raise ValueError(
            "all durations for split must be real non-negative number or "
            "set to -1 (meaning 'use the remaining dataset)"
        )

    if sum([split_dur == -1 for split_dur in split_durs.values()]) > 1:
        raise ValueError(
            "cannot specify duration of more than one split as -1, unclear how to calculate durations of splits."
        )

    # set any None values for durations to 0; no part of dataset will go to that split
    for split_name in split_durs.keys():
        if split_durs[split_name] is None:
            split_durs[split_name] = 0

    if -1 in split_durs.values():
        total_other_splits_dur = sum([dur for dur in split_durs.values() if dur != -1])

        if total_other_splits_dur > dataset_dur:
            raise ValueError(
                "One dataset split duration was specified as -1, but the total of the other durations specified, "
                f"{total_other_splits_dur} s, is greater than total duration of Dataset, {dataset_dur}."
            )
    else:  # if none of the target durations are -1
        total_splits_dur = sum(split_durs.values())

        if total_splits_dur > dataset_dur:
            raise ValueError(
                f"Total of the split durations specified, {total_splits_dur} s, "
                f"is greater than total duration of dataset, {dataset_dur}."
            )

    return split_durs["train"], split_durs["val"], split_durs["test"]



"""tests for vak.core.eval module"""
import pytest

import vak.config
import vak.constants
import vak.paths
import vak.core.eval


# written as separate function so we can re-use in tests/unit/test_cli/test_eval.py
def eval_output_matches_expected(model_config_map, output_dir):
    for model_name in model_config_map.keys():
        eval_csv = sorted(output_dir.glob(f"eval_{model_name}*csv"))
        assert len(eval_csv) == 1

    return True


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
    ],
)
def test_eval(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "EVAL", "option": "output_dir", "value": str(output_dir)},
        {"section": "EVAL", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="eval",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.eval.models)

    vak.core.eval(
        cfg.eval.csv_path,
        model_config_map,
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        window_size=cfg.dataloader.window_size,
        num_workers=cfg.eval.num_workers,
        spect_scaler_path=cfg.eval.spect_scaler_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        device=cfg.eval.device,
        logger=None,
    )

    assert eval_output_matches_expected(model_config_map, output_dir)



"""function that converts a set of array files (.npz, .mat) containing spectrograms
into a pandas DataFrame that represents a dataset used by ``vak``

the returned DataFrame has columns as specified by vak.io.spect.DF_COLUMNS
"""
from glob import glob
import os
from pathlib import Path

import dask.bag as db
from dask.diagnostics import ProgressBar
import numpy as np
import pandas as pd

from .. import constants
from .. import files
from ..annotation import source_annot_map
from ..converters import labelset_to_set
from ..logging import log_or_print


# constant, used for names of columns in DataFrame below
DF_COLUMNS = [
    "audio_path",
    "spect_path",
    "annot_path",
    "annot_format",
    "duration",
    "timebin_dur",
]


def to_dataframe(
    spect_format,
    spect_dir=None,
    spect_files=None,
    annot_list=None,
    annot_format=None,
    spect_annot_map=None,
    labelset=None,
    n_decimals_trunc=5,
    freqbins_key="f",
    timebins_key="t",
    spect_key="s",
    audio_path_key="audio_path",
    logger=None,
):
    """convert spectrogram files into a dataset of vocalizations represented as a Pandas DataFrame.
    Spectrogram files are array in .npz files created by numpy or in .mat files created by Matlab.

    Parameters
    ----------
    spect_format : str
        format of files containing spectrograms. One of {'mat', 'npz'}
    spect_dir : str
        path to directory of files containing spectrograms as arrays.
        Default is None.
    spect_files : list
        List of paths to array files. Default is None.
    annot_list : list
        of annotations for array files. Default is None
    annot_format : str
        name of annotation format. Added as a column to the DataFrame if specified.
        Used by other functions that open annotation files via their paths from the DataFrame.
        Should be a format that the crowsetta library recognizes.
        Default is None.
    spect_annot_map : dict
        Where keys are paths to files and value corresponding to each key is
        the annotation for that file.
        Default is None.
    labelset : str, list, set
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, then files will be skipped where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration calculated from
        the vector of time bins.
        Default is 3, i.e. assumes milliseconds is the last significant digit.
    freqbins_key : str
        key for accessing vector of frequency bins in files. Default is 'f'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    audio_path_key : str
        key for accessing path to source audio file for spectogram in files.
        Default is 'audio_path'.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    vak_df : pandas.Dataframe
        that represents a dataset of vocalizations.

    Notes
    -----
    Each file should contain a spectrogram as a matrix and two vectors associated with it, a
    vector of frequency bins and time bins, where the values in those vectors are the values
    at the bin centers. (As far as vak is concerned, "vector" and "matrix" are synonymous with
    "array".)

    Since both .mat files and .npz files load into a dictionary-like structure,
    the arrays will be accessed with keys. By convention, these keys are 's', 'f', and 't'.
    If you use different keys you can let this function know by changing
    the appropriate arguments: spect_key, freqbins_key, timebins_key
    """
    # pre-conditions ---------------------------------------------------------------------------------------------------
    if spect_format not in constants.VALID_SPECT_FORMATS:
        raise ValueError(
            f"spect_format must be one of '{constants.VALID_SPECT_FORMATS}'; "
            f"format '{spect_format}' not recognized."
        )

    if all([arg is None for arg in (spect_dir, spect_files, spect_annot_map)]):
        raise ValueError("must specify one of: spect_dir, spect_files, spect_annot_map")

    if spect_dir and spect_files:
        raise ValueError(
            "received values for spect_dir and spect_files, unclear which to use"
        )

    if spect_dir and spect_annot_map:
        raise ValueError(
            "received values for spect_dir and spect_annot_map, unclear which to use"
        )

    if spect_files and spect_annot_map:
        raise ValueError(
            "received values for spect_files and spect_annot_map, unclear which to use"
        )

    if annot_list and spect_annot_map:
        raise ValueError(
            "received values for annot_list and spect_annot_map, unclear which annotations to use"
        )

    if (annot_list or spect_annot_map) and (annot_format is None):
        if annot_list:
            raise ValueError(
                "an annot_list was provided, but no annot_format was specified"
            )
        elif spect_annot_map:
            raise ValueError(
                "a spect_annot_map was provided, but no annot_format was specified"
            )

    if annot_format is not None and (annot_list is None and spect_annot_map is None):
        raise ValueError(
            "an annot_format was specified but no annot_list or spect_annot_map was provided"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    # ---- get a list of spectrogram files + associated annotation files -----------------------------------------------
    if spect_dir:  # then get spect_files from that dir
        # note we already validated format above
        spect_files = glob(os.path.join(spect_dir, f"*{spect_format}"))

    if spect_files:  # (or if we just got them from spect_dir)
        if annot_list:
            spect_annot_map = source_annot_map(spect_files, annot_list)
        else:
            # no annotation, so map spectrogram files to None
            spect_annot_map = dict((spect_path, None) for spect_path in spect_files)

    # --- filter by labelset -------------------------------------------------------------------------------------------
    if (
        labelset
    ):  # then assume user wants to filter out files where annotation has labels not in labelset
        for spect_path in list(
            spect_annot_map.keys()
        ):  # iterate over keys so we can pop from dict without RuntimeError
            annot = spect_annot_map[spect_path]
            labels_set = set(annot.seq.labels)
            # below, set(labels_mapping) is a set of that dict's keys
            if not labels_set.issubset(set(labelset)):
                extra_labels = labels_set - set(labelset)
                # because there's some label in labels
                # that's not in labels_mapping
                log_or_print(
                    f"Found labels, {extra_labels}, in {Path(spect_path).name}, "
                    "that are not in labels_mapping. Skipping file.",
                    logger=logger,
                    level="info",
                )
                spect_annot_map.pop(spect_path)
                continue

    # ---- validate set of spectrogram files ---------------------------------------------------------------------------
    # regardless of whether we just made it or user supplied it
    spect_paths = list(spect_annot_map.keys())
    files.spect.is_valid_set_of_spect_files(
        spect_paths,
        spect_format,
        freqbins_key,
        timebins_key,
        spect_key,
        n_decimals_trunc,
        logger=logger,
    )

    # now that we have validated that duration of time bins is consistent across files, we can just open one file
    # to get that time bin duration. This way validation function has no side effects, like returning time bin, and
    # this is still relatively fast compared to looping through all files again
    timebin_dur = files.spect.timebin_dur(
        spect_paths[0], spect_format, timebins_key, n_decimals_trunc
    )

    # ---- actually make the dataframe ---------------------------------------------------------------------------------
    # this is defined here so all other arguments to 'to_dataframe' are in scope
    def _to_record(spect_annot_tuple):
        """helper function that enables parallelized creation of "records",
        i.e. rows for dataframe, from .
        Accepts a two-element tuple containing (1) a dictionary that represents a spectrogram
        and (2) annotation for that file"""
        spect_path, annot = spect_annot_tuple
        spect_dict = files.spect.load(spect_path, spect_format)

        spect_dur = spect_dict[spect_key].shape[-1] * timebin_dur
        if audio_path_key in spect_dict:
            audio_path = spect_dict[audio_path_key]
            if type(audio_path) == np.ndarray:
                # (because everything stored in .npz has to be in an ndarray)
                audio_path = audio_path.tolist()
        else:
            # try to figure out audio filename programmatically
            # if we can't, then we'll get back a None
            # (or an error)
            audio_path = files.spect.find_audio_fname(spect_path)

        if annot is not None:
            annot_path = annot.annot_path
        else:
            annot_path = np.nan

        def abspath(a_path):
            if isinstance(a_path, str) or isinstance(a_path, Path):
                return str(Path(a_path).absolute())
            elif np.isnan(a_path):
                return a_path

        record = tuple(
            [
                abspath(audio_path),
                abspath(spect_path),
                abspath(annot_path),
                annot_format if annot_format else constants.NO_ANNOTATION_FORMAT,
                spect_dur,
                timebin_dur,
            ]
        )
        return record

    spect_path_annot_tuples = db.from_sequence(spect_annot_map.items())
    log_or_print(
        "creating pandas.DataFrame representing dataset from spectrogram files",
        logger=logger,
        level="info",
    )
    with ProgressBar():
        records = list(spect_path_annot_tuples.map(_to_record))

    return pd.DataFrame.from_records(data=records, columns=DF_COLUMNS)



"""functions for plotting spectrograms"""
import matplotlib.pyplot as plt

from .annot import annotation


def spect(s, t, f, tlim=None, flim=None, ax=None, imshow_kwargs=None):
    """plot spectrogram

    Parameters
    ----------
    s : numpy.ndarray
        spectrogram, a matrix
    t : numpy.ndarray
        vector of centers of time bins from spectrogram
    f : numpy.ndarray
        vector of centers of frequency bins from spectrogram
    tlim : tuple, list
        limits of time axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of t will be plotted.
    flim : tuple, list
        limits of frequency axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of f will be plotted.
        limits of time axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of t will be plotted.
    flim : tuple, list
        limits of frequency axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of f will be plotted.
    ax : matplotlib.axes.Axes
        axes on which to plot spectrgraom
    imshow_kwargs : dict
        keyword arguments passed to matplotlib.axes.Axes.imshow method
        used to plot spectrogram. Default is None.
    """
    if imshow_kwargs is None:
        imshow_kwargs = {}

    if ax is None:
        fig, ax = plt.subplots()

    extent = [t.min(), t.max(), f.min(), f.max()]

    ax.imshow(s, aspect="auto", origin="lower", extent=extent, **imshow_kwargs)

    if tlim is not None:
        ax.set_xlim(tlim)

    if flim is not None:
        ax.set_ylim(flim)


def spect_annot(
    s,
    t,
    f,
    annot,
    tlim=None,
    flim=None,
    figsize=(20, 7.5),
    imshow_kwargs=None,
    line_kwargs=None,
    text_kwargs=None,
):
    """plot a spectrogram with annotated segments below it.
    Convenience function that calls `vak.plot.spect` and `vak.plot.annotation`

    Parameters
    ----------
    s : numpy.ndarray
        spectrogram, a matrix
    t : numpy.ndarray
        vector of centers of time bins from spectrogram
    f : numpy.ndarray
        vector of centers of frequency bins from spectrogram
    annot : crowsetta.Annotation
        annotation that has segments to be plotted
        (the `annot.seq.segments` attribute)    figsize
    tlim : tuple, list
        limits of time axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of t will be plotted.
    flim : tuple, list
        limits of frequency axis (min, max) (i.e., x-axis).
        Default is None, in which case entire range of f will be plotted.
    figsize : tuple, list
        figure size (width, height) in inches. Default is (20, 7.5).
    imshow_kwargs : dict
        keyword arguments that will get passed to `matplotlib.axes.Axes.imshow`
        when using that method to plot spectrogram.
    line_kwargs : dict
        keyword arguments for `LineCollection`.
        Passed to the function `vak.plot.annot.segments` that plots segments
        as a `LineCollection` instance. Default is None.
    text_kwargs : dict
        keyword arguments for `matplotlib.axes.Axes.text`.
        Passed to the function `vak.plot.annot.labels` that plots labels
        using Axes.text method.
        Defaults are defined as `vak.plot.annot.DEFAULT_TEXT_KWARGS`.

    Returns
    -------
    fig, spect_ax, annot_ax :
        matplotlib Figure and Axes instances.
        The spect_ax is the axes containing the spectrogram
        and the annot_ax is the axes containing the
        annotated segments.
    """
    fig = plt.figure(figsize=figsize)
    gs = fig.add_gridspec(3, 3)
    spect_ax = fig.add_subplot(gs[:2, :])
    annot_ax = fig.add_subplot(gs[2, :])

    spect(s, t, f, tlim, flim, ax=spect_ax, imshow_kwargs=imshow_kwargs)

    annotation(
        annot, t, tlim, ax=annot_ax, line_kwargs=line_kwargs, text_kwargs=text_kwargs
    )

    return fig, spect_ax, annot_ax



import time
from datetime import datetime

import vak.constants
import vak.timenow


def test_timenow():
    before_timestamp = time.mktime(datetime.now().timetuple())
    timenow_str = vak.timenow.get_timenow_as_str()
    assert len(timenow_str) == len(vak.constants.STRFTIME_TIMESTAMP)
    timenow_str_as_timestamp = time.mktime(
        datetime.strptime(timenow_str, vak.constants.STRFTIME_TIMESTAMP).timetuple()
    )
    after_timestamp = time.mktime(datetime.now().timetuple())
    assert before_timestamp <= timenow_str_as_timestamp <= after_timestamp



"""module that contains helper function to load models

Models in separate packages should make themselves available to vak by including
'vak.models' in the entry_points of their setup.py file.

For example, if you had a package `grunet` containing a model
that was instantiated by a function `GRUnet`,
then that package would include the following in its setup.py file:

setup(
    ...
    entry_points={'vak.models': 'GRUnet = grunet:GRUnet'},
    ...
)

For more detail on entry points in Python, see:
https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata
https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins
https://amir.rachum.com/blog/2017/07/28/python-entry-points/
"""
from .. import entry_points

MODELS_ENTRY_POINT = "vak.models"


def find():
    """find installed vak.models

    returns generator that yields model name and function for loading
    """
    for entrypoint in entry_points._iter(MODELS_ENTRY_POINT):
        yield entrypoint.name, entrypoint.load()


def from_model_config_map(model_config_map, num_classes, input_shape, logger=None):
    """get models that are ready to train, given their names and configurations.

    Given a dictionary that maps model names to configurations,
    along with the number of classes they should be trained to discriminate and their input shape,
    return a dictionary that maps model names to instances of the model

    Parameters
    ----------
    model_config_map : dict
        where each key-value pair is model name : dict of config parameters
    num_classes : int
        number of classes model will be trained to classify
    input_shape : tuple, list
        e.g. (channels, height, width).
        Batch size is not required for input shape.
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    models_map : dict
        where keys are model names and values are instances of the models, ready for training
    """
    MODELS = {model_name: model_builder for model_name, model_builder in find()}

    models_map = {}
    for model_name, model_config in model_config_map.items():
        # pass section dict as kwargs to config parser function
        model_config["network"].update(
            num_classes=num_classes,
            input_shape=input_shape,
        )
        try:
            model = MODELS[model_name].from_config(config=model_config)
        except KeyError:
            model = MODELS[f"{model_name}Model"].from_config(
                config=model_config, logger=logger
            )
        models_map[model_name] = model
    return models_map



"""tests for vak.cli.prep module"""
from pathlib import Path

import pytest

import vak.config
import vak.constants
import vak.core.train
import vak.paths
import vak.io.spect

from . import cli_asserts


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("predict", "cbin", None, "notmat"),
        ("predict", "wav", None, "birdsong-recognition-dataset"),
        ("train", "cbin", None, "notmat"),
        ("train", "wav", None, "birdsong-recognition-dataset"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_purpose_from_toml(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    specific_config,
    default_model,
    tmp_path,
):
    toml_path = specific_config(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
    )
    config_toml = vak.config.parse._load_toml_from_path(toml_path)
    vak.cli.prep.purpose_from_toml(config_toml)


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("predict", "cbin", None, "notmat"),
        ("predict", "wav", None, "birdsong-recognition-dataset"),
        ("train", "cbin", None, "notmat"),
        ("train", "wav", None, "birdsong-recognition-dataset"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_prep(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    specific_config,
    default_model,
    tmp_path,
):
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "PREP", "option": "output_dir", "value": str(output_dir)},
        # need to remove csv_path option from configs we already ran prep on to avoid error
        {
            "section": config_type.upper(),
            "option": "csv_path",
            "value": None,
        },
    ]
    toml_path = specific_config(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    vak.cli.prep.prep(toml_path)

    cfg = vak.config.parse.from_toml_path(toml_path)
    command_section = getattr(cfg, config_type)
    csv_path = getattr(command_section, "csv_path")
    # we don't bother checking whether csv is as expected
    # because that's already tested by `test_io.test_spect`, `test_io.test_dataframe`, etc.
    assert Path(csv_path).exists()

    assert cli_asserts.log_file_created(command="prep", output_path=cfg.prep.output_dir)


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("predict", "cbin", None, "notmat"),
        ("predict", "wav", None, "birdsong-recognition-dataset"),
        ("train", "cbin", None, "notmat"),
        ("train", "wav", None, "birdsong-recognition-dataset"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_prep_csv_path_raises(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    specific_config,
    default_model,
    tmp_path,
):
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "PREP", "option": "output_dir", "value": str(output_dir)},
    ]
    toml_path = specific_config(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    with pytest.raises(ValueError):
        vak.cli.prep.prep(toml_path)



"""tests for vak.cli.learncurve module"""
import pytest

import vak.config
import vak.constants
import vak.cli.learncurve

from . import cli_asserts
from ..test_core.test_learncurve import learncurve_output_matches_expected


def test_learncurve(specific_config, tmp_path, model, device):
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    vak.cli.learncurve.learning_curve(toml_path)

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.learncurve.models)
    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert learncurve_output_matches_expected(cfg, model_config_map, results_path)

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="learncurve", output_path=results_path)


@pytest.mark.parametrize("window_size", [44, 88, 176])
def test_learncurve_previous_run_path(
    specific_config, tmp_path, model, device, previous_run_path_factory, window_size
):
    root_results_dir = tmp_path.joinpath("test_learncurve_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
        {
            "section": "LEARNCURVE",
            "option": "previous_run_path",
            "value": str(previous_run_path_factory(model)),
        },
        {"section": "DATALOADER", "option": "window_size", "value": window_size},
    ]

    toml_path = specific_config(
        config_type="learncurve",
        model=model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )

    vak.cli.learncurve.learning_curve(toml_path)

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.learncurve.models)
    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert learncurve_output_matches_expected(cfg, model_config_map, results_path)

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="learncurve", output_path=results_path)



import pandas as pd

from . import annotation, files, labels, labeled_timebins


def has_unlabeled(csv_path, labelset, timebins_key="t"):
    """determine if a dataset has segments that are unlabeled.

    Used to decide whether an additional class needs to be added
    to the set of labels Y = {y_1, y_2, ... y_n}, where the added
    class y_n+1 will represent the unlabeled segments.

    Parameters
    ----------
    csv_path : str, Path
        to .csv file representing dataset.
    labelset : set
        of labels, str or int, e.g. {'a', 'b', 'c'}
    timebins_key : str
        key used to access timebins vector in spectrogram files.
        Default is 't'.

    Returns
    -------
    has_unlabeled : bool
        if True, dataset has unlabeled segments.
    """
    tmp_labelmap = labels.to_map(labelset, map_unlabeled=False)

    vak_df = pd.read_csv(csv_path)
    annots = annotation.from_df(vak_df)

    has_unlabeled_list = []
    for annot, spect_path in zip(annots, vak_df["spect_path"].values):
        time_bins = files.spect.load(spect_path)[timebins_key]
        lbls_int = [tmp_labelmap[lbl] for lbl in annot.seq.labels]
        has_unlabeled_list.append(
            labeled_timebins.has_unlabeled(
                lbls_int, annot.seq.onsets_s, annot.seq.offsets_s, time_bins
            )
        )

    return any(has_unlabeled_list)



import pytest

import vak.annotation
import vak.io.audio


def test_files_from_dir(annot_dir_notmat, annot_files_notmat):
    annot_files_from_dir = vak.annotation.files_from_dir(
        annot_dir_notmat, annot_format="notmat"
    )

    annot_files_notmat = [str(annot_file) for annot_file in annot_files_notmat]
    assert sorted(annot_files_from_dir) == sorted(annot_files_notmat)


@pytest.mark.parametrize(
    "source_type, source_format, annot_format",
    [
        ("audio", "cbin", "notmat"),
        ("audio", "wav", "birdsong-recognition-dataset"),
        ("spect", "mat", "yarden"),
    ],
)
def test_source_annot_map(
    source_type,
    source_format,
    annot_format,
    audio_list_factory,
    spect_list_mat,
    specific_annot_list,
):
    if source_type == "audio":
        source_files = audio_list_factory(source_format)
    else:
        source_files = spect_list_mat
    annot_list = specific_annot_list(annot_format)
    source_annot_map = vak.annotation.source_annot_map(
        source_files=source_files, annot_list=annot_list
    )

    # test all the audio paths made it into the map
    source_files_from_map = list(source_annot_map.keys())
    for source_file in source_files:
        assert source_file in source_files_from_map

    # test all the annots made it into the map
    annot_list_from_map = list(source_annot_map.values())
    for annot in annot_list:
        assert annot in annot_list_from_map

    # test all mappings are correct
    for source_path, annot in list(source_annot_map.items()):
        assert vak.annotation.recursive_stem(
            annot.audio_path
        ) == vak.annotation.recursive_stem(source_path)



"""sub-package that parses config.toml files and returns config object"""
from . import (
    config,
    dataloader,
    eval,
    learncurve,
    models,
    parse,
    predict,
    prep,
    spect_params,
    train,
    validators,
)



from datetime import datetime
from pathlib import Path
import warnings

from .. import split
from ..converters import expanded_user_path, labelset_to_set
from ..io import dataframe
from ..logging import log_or_print


VALID_PURPOSES = frozenset(
    [
        "eval",
        "learncurve",
        "predict",
        "train",
    ]
)


def prep(
    data_dir,
    purpose,
    output_dir=None,
    audio_format=None,
    spect_format=None,
    spect_params=None,
    spect_output_dir=None,
    annot_format=None,
    annot_file=None,
    labelset=None,
    train_dur=None,
    val_dur=None,
    test_dur=None,
    logger=None,
):
    """prepare datasets from vocalizations.
    High-level function that prepares datasets to be used by other
    high-level functions like vak.train, vak.predict, and vak.learncurve

    Parameters
    ----------
    data_dir : str, Path
        path to directory with files from which to make dataset
    purpose : str
        one of {'train', 'predict', 'learncurve'}
    output_dir : str
        Path to location where data sets should be saved.
        Default is None, in which case data sets to `data_dir`.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
        Default is None, but either audio_format or spect_format must be specified.
    spect_format : str
        format of files containing spectrograms as 2-d matrices. One of {'mat', 'npz'}.
        Default is None, but either audio_format or spect_format must be specified.
    spect_params : dict, vak.config.SpectParams
        parameters for creating spectrograms. Default is None.
    spect_output_dir : str
        path to location where spectrogram files should be saved.
        Default is None, in which case it defaults to ``data_dir``.
        A new directory will be created in ``spect_output_dir`` with
        the name 'spectrograms_generated_{time stamp}'.
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid. Default is None.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    labelset : str, list, set
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, then files will be skipped where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.
    train_dur : float
        total duration of training set, in seconds. When creating a learning curve,
        training subsets of shorter duration will be drawn from this set. Default is None.
    val_dur : float
        total duration of validation set, in seconds. Default is None.
    test_dur : float
        total duration of test set, in seconds. Default is None.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    vak_df : pandas.DataFrame
        that represents a dataset of vocalizations
    csv_path : Path
        to csv saved from vak_df

    Notes
    -----
    Saves a .csv file representing the dataset generated from data_dir.

    Datasets are used to train neural networks that segment audio files into
    vocalizations, and then predict labels for those segments.
    The function also prepares datasets so neural networks can predict the
    segmentation and annotation of vocalizations in them.
    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.

    If the 'purpose' is set to 'train' or 'learncurve', and/or
    the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    A duration can also be specified for a validation set
    (used to measure performance during training).
    In these cases, the 'split' column in the .csv
    identifies which files (rows) belong to the training, test, and
    validation sets created from that Dataset.

    If the 'purpose' is set to 'predict' or 'eval',
    or no durations for any of the training sets are specified,
    then the function assumes all the vocalizations constitute a single
    dataset, and for all rows the 'split' columns for that dataset
    will be 'predict' or 'test' (respectively).
    """
    # pre-conditions ---------------------------------------------------------------------------------------------------
    if purpose not in VALID_PURPOSES:
        raise ValueError(
            f"purpose must be one of: {VALID_PURPOSES}\nValue for purpose was: {purpose}"
        )

    if audio_format is None and spect_format is None:
        raise ValueError("Must specify either audio_format or spect_format")

    if audio_format and spect_format:
        raise ValueError(
            "Cannot specify both audio_format and spect_format, "
            "unclear whether to create spectrograms from audio files or "
            "use already-generated spectrograms from array files"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(f"data_dir not found: {data_dir}")

    if output_dir:
        output_dir = expanded_user_path(output_dir)
    else:
        output_dir = data_dir

    if not output_dir.is_dir():
        raise NotADirectoryError(f"output_dir not found: {output_dir}")

    if spect_output_dir:
        spect_output_dir = expanded_user_path(spect_output_dir)
        if not spect_output_dir.is_dir():
            raise NotADirectoryError(f"spect_output_dir not found: {spect_output_dir}")

    if purpose == "predict":
        if labelset is not None:
            warnings.warn(
                "purpose set to 'predict', but a labelset was provided."
                "This would cause an error because the dataframe.from_files section will attempt to "
                f"check whether the files in the data_dir ({data_dir}) have labels in "
                "labelset, even though those files don't have annotation.\n"
                "Setting labelset to None."
            )
            labelset = None
    log_or_print(msg=f"purpose for dataset: {purpose}", logger=logger, level="info")
    # ---- figure out file name ----------------------------------------------------------------------------------------
    data_dir_name = data_dir.name
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")
    csv_fname_stem = f"{data_dir_name}_prep_{timenow}"
    csv_path = output_dir.joinpath(f"{csv_fname_stem}.csv")

    # ---- figure out if we're going to split into train / val / test sets ---------------------------------------------
    # catch case where user specified duration for just training set, raise a helpful error instead of failing silently
    if (purpose == "train" or purpose == "learncurve") and (
        (train_dur is not None and train_dur > 0)
        and (val_dur is None or val_dur == 0)
        and (test_dur is None or val_dur == 0)
    ):
        raise ValueError(
            "duration specified for just training set, but prep function does not currently support creating a "
            "single split of a specified duration. Either remove the train_dur option from the prep section and "
            "rerun, in which case all data will be included in the training set, or specify values greater than "
            "zero for test_dur (and val_dur, if a validation set will be used)"
        )

    if all([dur is None for dur in (train_dur, val_dur, test_dur)]) or purpose in (
        "eval",
        "predict",
    ):
        # then we're not going to split
        log_or_print(msg="will not split dataset", logger=logger, level="info")
        do_split = False
    else:
        if val_dur is not None and train_dur is None and test_dur is None:
            raise ValueError(
                "cannot specify only val_dur, unclear how to split dataset into training and test sets"
            )
        else:
            log_or_print(msg="will split dataset", logger=logger, level="info")
            do_split = True

    # ---- actually make the dataset -----------------------------------------------------------------------------------
    vak_df = dataframe.from_files(
        labelset=labelset,
        data_dir=data_dir,
        annot_format=annot_format,
        annot_file=annot_file,
        audio_format=audio_format,
        spect_format=spect_format,
        spect_output_dir=spect_output_dir,
        spect_params=spect_params,
        logger=logger,
    )

    if do_split:
        # save before splitting, jic duration args are not valid (we can't know until we make dataset)
        vak_df.to_csv(csv_path)
        vak_df = split.dataframe(
            vak_df,
            labelset=labelset,
            train_dur=train_dur,
            val_dur=val_dur,
            test_dur=test_dur,
            logger=logger,
        )

    elif (
        do_split is False
    ):  # add a split column, but assign everything to the same 'split'
        # ideally we would just say split=purpose in call to add_split_col, but
        # we have to special case, because "eval" looks for a 'test' split (not an "eval" split)
        if purpose == "eval":
            split_name = "test"  # 'split_name' to avoid name clash with split package
        elif purpose == "predict":
            split_name = "predict"

        vak_df = dataframe.add_split_col(vak_df, split=split_name)

    log_or_print(
        msg=f"saving dataset as a .csv file: {csv_path}", logger=logger, level="info"
    )
    vak_df.to_csv(
        csv_path, index=False
    )  # index is False to avoid having "Unnamed: 0" column when loading

    return vak_df, csv_path



from datetime import datetime

from .constants import STRFTIME_TIMESTAMP


def get_timenow_as_str():
    f"""returns current time as a string,
    with the format specified by ``vak.constants.STRFTIME_TIMESTAMP``"""
    return datetime.now().strftime(STRFTIME_TIMESTAMP)



from datetime import datetime
from pathlib import Path

from .. import config
from .. import core
from .. import logging


def eval(toml_path):
    """evaluate a trained model with dataset specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml_path(toml_path)

    if cfg.eval is None:
        raise ValueError(
            f"eval called with a config.toml file that does not have a EVAL section: {toml_path}"
        )

    # ---- set up logging ----------------------------------------------------------------------------------------------
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")
    logger = logging.get_logger(
        log_dst=cfg.eval.output_dir,
        caller="eval",
        timestamp=timenow,
        logger_name=__name__,
    )
    logger.info("Logging results to {}".format(cfg.eval.output_dir))

    model_config_map = config.models.map_from_path(toml_path, cfg.eval.models)

    core.eval(
        cfg.eval.csv_path,
        model_config_map,
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        window_size=cfg.dataloader.window_size,
        num_workers=cfg.eval.num_workers,
        spect_scaler_path=cfg.eval.spect_scaler_path,
        spect_key=cfg.spect_params.spect_key,
        timebins_key=cfg.spect_params.timebins_key,
        device=cfg.eval.device,
        logger=logger,
    )



from pathlib import Path

import toml
from toml.decoder import TomlDecodeError

from .config import Config
from .dataloader import DataLoaderConfig
from .eval import EvalConfig
from .learncurve import LearncurveConfig
from .predict import PredictConfig
from .prep import PrepConfig
from .spect_params import SpectParamsConfig
from .train import TrainConfig
from .validators import are_sections_valid, are_options_valid

SECTION_CLASSES = {
    "DATALOADER": DataLoaderConfig,
    "EVAL": EvalConfig,
    "LEARNCURVE": LearncurveConfig,
    "PREDICT": PredictConfig,
    "PREP": PrepConfig,
    "SPECT_PARAMS": SpectParamsConfig,
    "TRAIN": TrainConfig,
}

REQUIRED_OPTIONS = {
    "DATALOADER": None,
    "EVAL": [
        "checkpoint_path",
        "labelmap_path",
        "output_dir",
        "models",
    ],
    "LEARNCURVE": [
        "models",
        "root_results_dir",
        "train_set_durs",
        "num_replicates",
    ],
    "PREDICT": [
        "checkpoint_path",
        "labelmap_path",
        "models",
    ],
    "PREP": [
        "data_dir",
        "output_dir",
    ],
    "SPECT_PARAMS": None,
    "TRAIN": [
        "models",
        "root_results_dir",
    ],
}


def parse_config_section(config_toml, section_name, toml_path=None):
    """parse section of config.toml file

    Parameters
    ----------
    config_toml : dict
        containing config.toml file already loaded by parse function
    section_name : str
        name of section from configuration
        file that should be parsed
    toml_path : str
        path to a configuration file in TOML format. Default is None.
        Used for error messages if specified.

    Returns
    -------
    config : vak.config section class
        instance of class that represents section of config.toml file,
        e.g. PredictConfig for 'PREDICT' section
    """
    section = dict(config_toml[section_name].items())

    required_options = REQUIRED_OPTIONS[section_name]
    if required_options is not None:
        for required_option in required_options:
            if required_option not in section:
                if toml_path:
                    err_msg = (
                        f"the '{required_option}' option is required but was not found in the "
                        f"{section_name} section of the config.toml file: {toml_path}"
                    )
                else:
                    err_msg = (
                        f"the '{required_option}' option is required but was not found in the "
                        f"{section_name} section of the toml config"
                    )
                raise KeyError(err_msg)
    return SECTION_CLASSES[section_name](**section)


def _validate_sections_arg_convert_list(sections):
    if isinstance(sections, str):
        sections = [sections]
    elif isinstance(sections, list):
        if not all([isinstance(section_name, str) for section_name in sections]):
            raise ValueError("all section names in 'sections' should be strings")
        if not all(
            [section_name in list(SECTION_CLASSES.keys()) for section_name in sections]
        ):
            raise ValueError(
                "all section names in 'sections' should be valid names of sections. "
                f"Values for 'sections were: {sections}.\n"
                f"Valid section names are: {list(SECTION_CLASSES.keys())}"
            )
    return sections


def from_toml(config_toml, toml_path=None, sections=None):
    """load a TOML configuration file

    Parameters
    ----------
    config_toml : dict
        Python ``dict`` containing a .toml configuration file,
        parsed by the ``toml`` library.
    toml_path : str, Path
        path to a configuration file in TOML format. Default is None.
        Not required, used only to make any error messages clearer.
    sections : str, list
        name of section or sections from configuration
        file that should be parsed. Can be a string
        (single section) or list of strings (multiple
        sections). Default is None,
        in which case all are validated and parsed.

    Returns
    -------
    config : vak.config.parse.Config
        instance of Config class, whose attributes correspond to
        sections in a config.toml file.
    """
    are_sections_valid(config_toml, toml_path)

    sections = _validate_sections_arg_convert_list(sections)

    config_dict = {}
    if sections is None:
        sections = list(
            SECTION_CLASSES.keys()
        )  # i.e., parse all sections, except model
    for section_name in sections:
        if section_name in config_toml:
            are_options_valid(config_toml, section_name, toml_path)
            config_dict[section_name.lower()] = parse_config_section(
                config_toml, section_name, toml_path
            )

    return Config(**config_dict)


def _load_toml_from_path(toml_path):
    """helper function to load toml config file,
    factored out to use in other modules when needed

    checks if ``toml_path`` exists before opening,
    and tries to give a clear message if an error occurs when parsing"""
    toml_path = Path(toml_path)
    if not toml_path.is_file():
        raise FileNotFoundError(f".toml config file not found: {toml_path}")

    try:
        with toml_path.open("r") as fp:
            config_toml = toml.load(fp)
    except TomlDecodeError as e:
        raise Exception(f"Error when parsing .toml config file: {toml_path}") from e

    return config_toml


def from_toml_path(toml_path, sections=None):
    """parse a TOML configuration file

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
        Parsed by ``toml`` library, then converted to an
        instance of ``vak.config.parse.Config`` by
        calling ``vak.parse.from_toml``
    sections : str, list
        name of section or sections from configuration
        file that should be parsed. Can be a string
        (single section) or list of strings (multiple
        sections). Default is None,
        in which case all are validated and parsed.

    Returns
    -------
    config : vak.config.parse.Config
        instance of Config class, whose attributes correspond to
        sections in a config.toml file.
    """
    config_toml = _load_toml_from_path(toml_path)
    return from_toml(config_toml, toml_path, sections)



import pytest

import vak.metrics.distance.functional


LEV_PARAMETRIZE = [
    # adapted from https://github.com/toastdriven/pylev/blob/master/tests.py
    ("kitten", "sitting", 3),
    ("kitten", "kitten", 0),
    ("", "", 0),
    ("kitten", "", 6),
    ("", "sitting", 7),
    ("meilenstein", "levenshtein", 4),
    ("levenshtein", "frankenstein", 6),
    ("confide", "deceit", 6),
    ("CUNsperrICY", "conspiracy", 8),
    # case added to catch failure with our previous implementation from
    # https://en.wikibooks.org/wiki/Talk:Algorithm_Implementation/Strings/Levenshtein_distance#Bug_in_vectorized_(5th)_version
    ('aabcc', 'bccdd', 4),
]


@pytest.mark.parametrize(
    "source, target, expected_distance",
    LEV_PARAMETRIZE
)
def test_levenshtein(source, target, expected_distance):
    distance = vak.metrics.distance.functional.levenshtein(source, target)
    assert distance == expected_distance


@pytest.mark.parametrize(
    "y_pred, y_true, expected_distance",
    LEV_PARAMETRIZE
)
def test_segment_error_rate(y_pred, y_true, expected_distance):
    if len(y_true) == 0 and len(y_pred) == 0:
        expected_rate = 0
        rate = vak.metrics.distance.functional.segment_error_rate(y_pred, y_true)
        assert rate == expected_rate
    elif len(y_true) == 0 and len(y_pred) != 0:
        with pytest.raises(ValueError):
            vak.metrics.distance.functional.segment_error_rate(y_pred, y_true)
    else:
        expected_rate = expected_distance / len(y_true)
        rate = vak.metrics.distance.functional.segment_error_rate(y_pred, y_true)
        assert rate == expected_rate



from math import isclose

import numpy as np
import pandas as pd
import pytest

import vak.io.spect
import vak.annotation
import vak.split.split


NUM_SAMPLES = 10  # number of times to sample behavior of random-number generator


def train_test_dur_split_inds_output_matches_expected(
    train_dur,
    val_dur,
    test_dur,
    labelset,
    durs,
    labels,
    train_inds,
    val_inds,
    test_inds,
):
    for split, dur_in, inds in zip(
        ("train", "val", "test"),
        (train_dur, val_dur, test_dur),
        (train_inds, val_inds, test_inds),
    ):
        if dur_in is not None:
            dur_out = sum([durs[ind] for ind in inds])
            if dur_in >= 0:
                assert dur_out >= dur_in
            elif dur_in == -1:
                if split == "train":
                    assert isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in test_inds])
                    )
                elif split == "test":
                    assert isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in train_inds])
                    )

            all_lbls_this_set = [lbl for ind in inds for lbl in labels[ind]]
            assert labelset == set(all_lbls_this_set)
        else:
            assert inds is None

    assert set(train_inds).isdisjoint(set(test_inds))
    if val_dur is not None:
        assert set(train_inds).isdisjoint(set(val_inds))
        assert set(test_inds).isdisjoint(set(val_inds))

    return True


@pytest.mark.parametrize(
    "durs, labels, labelset, train_dur, val_dur, test_dur",
    [
        (
            (5, 5, 5, 5, 5),
            ([np.asarray(list("abcde")) for _ in range(5)]),
            set(list("abcde")),
            20,
            None,
            5,
        ),
        (
            (3, 2, 1, 3, 2, 3, 2, 1, 3, 2),
            ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"],
            set(list("abcde")),
            14,
            None,
            8,
        ),
        (
            (3, 2, 1, 3, 2, 3, 2, 1, 3, 2),
            ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"],
            set(list("abcde")),
            8,
            7,
            7,
        ),
    ],
)
def test_train_test_dur_split_inds_fake_data(
    durs, labels, labelset, train_dur, val_dur, test_dur
):
    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_train_test_dur_split_inds_fake_data_impossible():
    durs = (3, 2, 1, 3, 2, 3, 2, 1, 3, 2)
    labelset = set(list("abcde"))
    labels = ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"]
    labels = [np.asarray(list(lbl)) for lbl in labels]
    train_dur = 16
    val_dur = 2
    test_dur = 4
    with pytest.raises(ValueError):
        vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


@pytest.mark.parametrize(
    "train_dur, val_dur, test_dur",
    [
        (35, 20, 35),
        (35, None, -1),
        (-1, None, 35),
    ],
)
def test_train_test_dur_split_inds_audio_cbin_annot_notmat(
    train_dur, val_dur, test_dur, audio_cbin_annot_notmat_durs_labels, labelset_notmat
):
    labelset_notmat = set(labelset_notmat)
    durs, labels = audio_cbin_annot_notmat_durs_labels

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset_notmat, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset_notmat,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


@pytest.mark.parametrize(
    "train_dur, val_dur, test_dur",
    [(200, 100, 200), (200, None, -1), (-1, None, 200), (200, 100, 200)],
)
def test_train_test_dur_split_inds_spect_mat(
    train_dur, val_dur, test_dur, spect_mat_annot_yarden_durs_labels, labelset_yarden
):
    labelset_yarden = set(labelset_yarden)
    durs, labels = spect_mat_annot_yarden_durs_labels
    train_dur = 200
    val_dur = 100
    test_dur = 200

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset_yarden, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset_yarden,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_dataframe_None_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = None
    val_dur = None
    test_dur = None

    with pytest.raises(ValueError):
        vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_only_val_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = None
    val_dur = 100
    test_dur = None

    # because we only specified duration for validation set
    with pytest.raises(ValueError):
        vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_negative_dur_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = -2
    test_dur = None
    val_dur = 100

    # because negative duration is invalid
    with pytest.raises(ValueError):
        vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_specd_dur_gt_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = 100
    test_dur = 100
    val_dur = 100
    # because total splits duration is greater than dataset duration
    with pytest.raises(ValueError):
        vak.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


@pytest.mark.parametrize("train_dur, test_dur", [(200, 200), (200, None), (None, 200)])
def test_split_dataframe_mat(
    train_dur, test_dur, spect_dir_mat, annot_list_yarden, labelset_yarden
):
    labelset_yarden = set(labelset_yarden)

    vak_df = vak.io.spect.to_dataframe(
        spect_format="mat",
        annot_format="yarden",
        spect_dir=spect_dir_mat,
        annot_list=annot_list_yarden,
    )

    train_dur = 200
    test_dur = 200

    vak_df_split = vak.split.split.dataframe(
        vak_df, labelset=labelset_yarden, train_dur=train_dur, test_dur=test_dur
    )

    assert isinstance(vak_df_split, pd.DataFrame)

    if train_dur is not None:
        train_dur_out = vak_df_split[vak_df_split["split"] == "train"].duration.sum()
        assert train_dur_out >= train_dur
    else:
        assert "train" not in vak_df_split["split"].unique().tolist()

    if test_dur is not None:
        test_dur_out = vak_df_split[vak_df_split["split"] == "test"].duration.sum()
        assert test_dur_out >= test_dur
    else:
        assert "test" not in vak_df_split["split"].unique().tolist()



from . import functional as F

__all__ = ["Levenshtein", "SegmentErrorRate"]


class Levenshtein:
    """levenshtein distance
    returns number of deletions, insertions, or substitutions
    required to convert source string into target string.

    Parameters
    ----------
    source, target : str

    Returns
    -------
    distance : int
        number of deletions, insertions, or substitutions
        required to convert source into target.
    """

    def __init__(self):
        pass

    def __call__(self, y_pred, y_true):
        return F.levenshtein(y_pred, y_true)


class SegmentErrorRate:
    """Levenshtein edit distance normalized by length of true sequence.
    Also known as word error distance; here applied to other vocalizations
    in addition to speech.

    Parameters
    ----------
    y_pred : str
        predicted labels for a series of songbird syllables
    y_true : str
        ground truth labels for a series of songbird syllables

    Returns
    -------
    Levenshtein distance(y_pred, y_true) / len(y_true)
    """

    def __init__(self):
        pass

    def __call__(self, y_pred, y_true):
        return F.segment_error_rate(y_pred, y_true)
